#ifndef DMLITE_DSI_GRIDFTP_H
#define DMLITE_DSI_GRIDFTP_H

#ifndef _LARGEFILE64_SOURCE
#define _LARGEFILE64_SOURCE
#endif

#include "globus_gridftp_server.h"

#include <dmlite/c/pool.h>
#include <dmlite/c/io.h>

extern globus_version_t dmlite_gridftp_local_version;

#define SUBJECT_MAX 256

/*
 * The handle structure (we store here what we need for our plugin).
 */
typedef struct globus_l_gfs_dmlite_handle_s {
	struct dmlite_manager *	manager;
	dmlite_fd *		fd;
	char			subject[SUBJECT_MAX];
	char			client_host[HOST_NAME_MAX];
	char			pfn[PATH_MAX];
	dmlite_location *	location;
	dmlite_any_dict *	location_extras;
	globus_bool_t		allow_local;
	globus_bool_t		is_replica;
	globus_mutex_t		gfs_mutex;
	globus_gfs_operation_t	gfs_op;
	int			optimal_count;
	globus_size_t		block_size; // Set using a globus call (amount of bytes to handle in each operation)
	// The two below are used for partial reads / writes (offset and length)
	globus_off_t		cur_length; // Keeps the total amount to read/write
	globus_off_t		cur_offset; // Keeps the current offset
	globus_result_t		cur_result;
	globus_bool_t		done;
	int			pending;
} globus_l_gfs_dmlite_handle_t;

typedef globus_l_gfs_dmlite_handle_t dmlite_handle_t;

/**
 * Functions called by globus (docs in the main source).
 */
int globus_l_gfs_dmlite_activate();

int globus_l_gfs_dmlite_deactivate();

void globus_l_gfs_dmlite_start(globus_gfs_operation_t op, globus_gfs_session_info_t * session_info);

void globus_l_gfs_dmlite_destroy(void * user_arg);

void globus_l_gfs_dmlite_command(globus_gfs_operation_t op, globus_gfs_command_info_t * cmd_info,
                                 void * user_arg);

void globus_l_gfs_dmlite_stat(globus_gfs_operation_t op, globus_gfs_stat_info_t * stat_info,
                              void * user_arg);

void globus_l_gfs_dmlite_send(globus_gfs_operation_t op, globus_gfs_transfer_info_t * transfer_info,
                              void * user_arg);

void globus_l_gfs_dmlite_recv(globus_gfs_operation_t op, globus_gfs_transfer_info_t * transfer_info,
                              void * user_arg);

#endif /* DMLITE_DSI_GRIDFTP_H */
