/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */

#include "dmlite_gridftp.h"
#include "dmlite_internal.h"

/**
 * Globals for this library.
 */
globus_version_t dmlite_gridftp_local_version =
{
	0, /* major version number */
	2, /* minor version number */
	1174988050,
	0 /* branch ID */
};

/*
 *  Interface definition for dmlite.
 */
static globus_gfs_storage_iface_t globus_l_gfs_dmlite_dsi_iface =
{
	GLOBUS_GFS_DSI_DESCRIPTOR_BLOCKING | GLOBUS_GFS_DSI_DESCRIPTOR_SENDER,
	globus_l_gfs_dmlite_start,
	globus_l_gfs_dmlite_destroy,
	NULL, /* list */
	globus_l_gfs_dmlite_send,
	globus_l_gfs_dmlite_recv,
	NULL, /* trev */
	NULL, /* active */
	NULL, /* passive */
	NULL, /* data destroy */
	globus_l_gfs_dmlite_command,
	globus_l_gfs_dmlite_stat,
	NULL,
	NULL
};

/*
 * Module definitions.
 *
 * Called when library loads.
 */
GlobusExtensionDefineModule(globus_gridftp_server_dmlite) =
{
	"globus_gridftp_server_dmlite",
	globus_l_gfs_dmlite_activate,
	globus_l_gfs_dmlite_deactivate,
	NULL,
	NULL,
	&dmlite_gridftp_local_version
};

/*
 * Called when the dmlite module is activated.
 */
int globus_l_gfs_dmlite_activate()
{
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_INFO, "activating 'dmlite' DSI module");
	globus_extension_registry_add(
		GLOBUS_GFS_DSI_REGISTRY,
		"dmlite",
		GlobusExtensionMyModule(globus_gridftp_server_dmlite),
		&globus_l_gfs_dmlite_dsi_iface);

	return 0;
}

/*
 * Called when the dmlite module is deactivated.
 */
int globus_l_gfs_dmlite_deactivate()
{
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_INFO, "deactivating 'dmlite' DSI module");
	globus_extension_registry_remove(
		GLOBUS_GFS_DSI_REGISTRY, "dmlite");

	return 0;
}

/*************************************************************************
 *  start
 *  -----
 *  This function is called when a new session is initialized, ie a user
 *  connectes to the server.  This hook gives the dsi an oppertunity to
 *  set internal state that will be threaded through to all other
 *  function calls associated with this session.  And an oppertunity to
 *  reject the user.
 *
 *  finished_info.info.session.session_arg should be set to an DSI
 *  defined data structure.  This pointer will be passed as the void *
 *  user_arg parameter to all other interface functions.
 *
 *  NOTE: at nice wrapper function should exist that hides the details
 *        of the finished_info structure, but it currently does not.
 *        The DSI developer should jsut follow this template for now
 ************************************************************************/
void globus_l_gfs_dmlite_start(
		globus_gfs_operation_t		op,
		globus_gfs_session_info_t *	session_info)
{
	globus_gfs_finished_info_t	finished_info;
	globus_result_t			result = GLOBUS_SUCCESS;
	dmlite_handle_t *		dmlite_handle_obj = NULL;
	char *				options;
	char *				nextop;
	char *				tok;
	char				config[PATH_MAX] = "/etc/dmlite.conf";

	GlobusGFSName(globus_l_gfs_dmlite_start);
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "starting new connection");
	memset(&finished_info, 0, sizeof(finished_info));

	/* Check for valid delegated credentials */
	if (!session_info->del_cred) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj,
				EFAULT, "could not find delegated credentials");
		goto out;
	}

	/* Allocate memory and initialize structs */
	dmlite_handle_obj = (dmlite_handle_t *)globus_malloc(sizeof(dmlite_handle_t));
	if (!dmlite_handle_obj) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj,
				EFAULT, "failed to allocate handle");
		goto out;
	}
	memset(dmlite_handle_obj, 0, sizeof(dmlite_handle_t));

	globus_mutex_init(&dmlite_handle_obj->gfs_mutex, GLOBUS_NULL);
	strncpy(dmlite_handle_obj->subject, session_info->subject, SUBJECT_MAX);
	dmlite_gfs_hostid2host(session_info->host_id, dmlite_handle_obj->client_host);

	globus_gridftp_server_get_config_string(op, &options);
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "DSI options: %s", options);
	nextop = options;
	while((tok = strsep(&nextop, ",")))
		if (!strcasecmp(tok, "allow_local")) {
			dmlite_handle_obj->allow_local = GLOBUS_TRUE;
		} else if (strcasestr(tok, "dmlite_config=") == tok) {
			strncpy(config, tok + 14, PATH_MAX);
		}
	globus_free(options);

	/* Initialize a dmlite manager */
	dmlite_handle_obj->manager = dmlite_manager_new();
	if (dmlite_manager_load_configuration(dmlite_handle_obj->manager, config)) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj,
				EFAULT, "failed to initialize manager :: %s",
				dmlite_manager_error(dmlite_handle_obj->manager));
	}

out:
	/* Some defaults for the globus return info */
	finished_info.type = GLOBUS_GFS_OP_SESSION_START;
	finished_info.info.session.session_arg = dmlite_handle_obj;
	finished_info.info.session.username = session_info->username;
	finished_info.info.session.home_dir = "/";
	finished_info.result = result;
	globus_gridftp_server_operation_finished(op, result, &finished_info);
}

/*************************************************************************
 *  destroy
 *  -------
 *  This is called when a session ends, ie client quits or disconnects.
 *  The dsi should clean up all memory they associated wit the session
 *  here. 
 ************************************************************************/
void globus_l_gfs_dmlite_destroy(void * user_arg)
{
	dmlite_handle_t *	dmlite_handle_obj = (dmlite_handle_t *)user_arg;

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "destroying dmlite handle");

	/* Now destroy whatever is still there */
	if(dmlite_handle_obj) {
		globus_mutex_destroy(&dmlite_handle_obj->gfs_mutex);
		if (dmlite_handle_obj->location)
			dmlite_location_free(dmlite_handle_obj->location);
		if (dmlite_handle_obj->manager)
			dmlite_manager_free(dmlite_handle_obj->manager);
		globus_free(dmlite_handle_obj);
	}
}

/*************************************************************************
 *  stat
 *  ----
 *  This interface function is called whenever the server needs 
 *  information about a given file or resource.  It is called then an
 *  LIST is sent by the client, when the server needs to verify that 
 *  a file exists and has the proper permissions, etc.
 ************************************************************************/
static void globus_l_gfs_dmlite_stat_free(
		globus_gfs_stat_t *	stat_array,
		int			stat_count)
{
	int i;

	for (i = 0; i < stat_count; i++) {
		if (stat_array[i].name) {
			globus_free(stat_array[i].name);
		}
		if (stat_array[i].symlink_target) {
			globus_free(stat_array[i].symlink_target);
		}
	}
	if (stat_array)
		globus_free(stat_array);
}

void globus_l_gfs_dmlite_stat(
		globus_gfs_operation_t		op,
		globus_gfs_stat_info_t *	stat_info,
		void *				user_arg)
{
	globus_result_t		result = GLOBUS_SUCCESS;
	globus_gfs_stat_t *	gfs_stat_array = NULL;
	int			stat_count = 0;
	dmlite_handle_t *	dmlite_handle_obj = (dmlite_handle_t *)user_arg;
	struct dmlite_context *	dmlite_context_obj;
	struct dmlite_xstat	dmlite_xstat_obj;
	struct dmlite_dir *	dir_fd;
	struct dmlite_xstat *	dir_entry_xstat;
	int			nfiles;
	const char *		localpath = dmlite_gfs_fixpath(stat_info->pathname, GLOBUS_FALSE);
	const char *		rfn = dmlite_gfs_fixpath(stat_info->pathname, GLOBUS_TRUE);

	GlobusGFSName(globus_l_gfs_dmlite_stat);

	dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_INFO, "stat :: %s", localpath);

	/* Get a dmlite_context object */
	dmlite_context_obj = dmlite_get_context(dmlite_handle_obj);
	if (!dmlite_context_obj) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj,
				EFAULT, "failed to create context");
		goto errout;
	}
	memset(&dmlite_xstat_obj, 0, sizeof(dmlite_xstat_obj));

	/* At first try to stat in virtual namespace */
	if (!dmlite_statx(dmlite_context_obj, localpath, &dmlite_xstat_obj))
		goto finstat;
	if (dmlite_errno(dmlite_context_obj) != ENOENT) {
		result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
		goto errout;
	}
	/* At second try to stat a local replica */
	if (!dmlite_rstatx(dmlite_context_obj, rfn, &dmlite_xstat_obj))
		goto finstat;
	if (dmlite_errno(dmlite_context_obj) != ENOENT) {
		result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
		goto errout;
	}
	/* Finally fall back to local file if permitted */
	if (!dmlite_handle_obj->allow_local) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, ENOENT, "local access is forbidden");
		goto errout;
	}
	if (stat(localpath, &dmlite_xstat_obj.stat)) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, errno, "failed to stat");
		goto errout;
	}
	strncpy(dmlite_xstat_obj.name, localpath, NAME_MAX);

finstat:
	/* File like info on the entry is easy to handle */
	if (!S_ISDIR(dmlite_xstat_obj.stat.st_mode) || stat_info->file_only) {
		dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "stat :: %s :: entry info only", localpath);
		/* Allocate the stat array */
		gfs_stat_array = (globus_gfs_stat_t *)globus_malloc(sizeof(globus_gfs_stat_t));
		if (!gfs_stat_array) {
			result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT, "failed to allocate array");
			goto errout;
		}
		dmlite_stat2gfs(dmlite_xstat_obj.name, &dmlite_xstat_obj.stat, gfs_stat_array);
		stat_count = 1;
	} else { /* Going through directory is a bit more work */
		nfiles  = dmlite_xstat_obj.stat.st_nlink;
		dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "stat :: %s :: full dir info with %d files", localpath, nfiles);

		/* Allocate the stat array */
		gfs_stat_array = (globus_gfs_stat_t *)globus_malloc(sizeof(globus_gfs_stat_t) * nfiles);
		if (!gfs_stat_array) {
			result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT, "failed to allocate array");
			goto errout;
		}

		/* Open the directory */
		dir_fd = dmlite_opendir(dmlite_context_obj, localpath);
		if (dir_fd == NULL) {
			result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
			goto errout;
		}

		while ((dir_entry_xstat = dmlite_readdirx(dmlite_context_obj, dir_fd))) {
			stat_count++;
			/* Double check that nlink value was correct and we don't overflow */
			if (stat_count > nfiles) {
				result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT,
					"we are reading file %d of %d (this should not happen)",
					stat_count, nfiles);
				goto errout;
			}
			dmlite_stat2gfs(dir_entry_xstat->name, &dir_entry_xstat->stat, gfs_stat_array + stat_count - 1);
		}

		/* Check for error */
		if (dmlite_errno(dmlite_context_obj) != 0) {
			result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
			goto errout;
		}

		/* Close the directory */
		if (dmlite_closedir(dmlite_context_obj, dir_fd)) {
			result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
			goto errout;
		}
	}

	globus_gridftp_server_finished_stat(op, GLOBUS_SUCCESS, gfs_stat_array, stat_count);
	globus_l_gfs_dmlite_stat_free(gfs_stat_array, stat_count);
	dmlite_context_free(dmlite_context_obj);
	return;

errout:
	globus_l_gfs_dmlite_stat_free(gfs_stat_array, stat_count);
	if (dmlite_context_obj)
		dmlite_context_free(dmlite_context_obj);
	if (result == GLOBUS_SUCCESS)
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT, "unknown error");
	globus_gridftp_server_finished_stat(op, result, GLOBUS_NULL, 0); // TODO: give a proper error here
}

/*************************************************************************
 *  command
 *  -------
 *  This interface function is called when the client sends a 'command'.
 *  commands are such things as mkdir, remdir, delete.  The complete
 *  enumeration is below.
 *
 *  To determine which command is being requested look at:
 *      cmd_info->command
 *
 *      GLOBUS_GFS_CMD_MKD = 1,
 *      GLOBUS_GFS_CMD_RMD,
 *      GLOBUS_GFS_CMD_DELE,
 *      GLOBUS_GFS_CMD_RNTO,
 *      GLOBUS_GFS_CMD_RNFR,
 *      GLOBUS_GFS_CMD_CKSM,
 *      GLOBUS_GFS_CMD_SITE_CHMOD,
 *      GLOBUS_GFS_CMD_SITE_DSI
 ************************************************************************/
void globus_l_gfs_dmlite_command(
		globus_gfs_operation_t		op,
		globus_gfs_command_info_t *	cmd_info,
		void *				user_arg)
{
	globus_result_t			result = GLOBUS_SUCCESS;
	dmlite_handle_t *		dmlite_handle_obj = (dmlite_handle_t *)user_arg;
	struct dmlite_context *		dmlite_context_obj;
	int				retval = 0;
	const char *			localpath = dmlite_gfs_fixpath(cmd_info->pathname, GLOBUS_FALSE);
	const char *			rfn = dmlite_gfs_fixpath(cmd_info->pathname, GLOBUS_TRUE);

	GlobusGFSName(globus_l_gfs_dmlite_command);

	dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_DUMP, "command :: %d :: %s",
		cmd_info->command, localpath);

	/* Get a dmlite_context object */
	dmlite_context_obj = dmlite_get_context(dmlite_handle_obj);
	if (!dmlite_context_obj) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT, "failed to create context");
		goto errout;
	}

	switch(cmd_info->command)
	{
		case GLOBUS_GFS_CMD_MKD: // mkdir
			retval = dmlite_mkdir(dmlite_context_obj, localpath, 0755);
			break;

		case GLOBUS_GFS_CMD_RMD: // rmdir
			retval = dmlite_rmdir(dmlite_context_obj, localpath);
			break;

		case GLOBUS_GFS_CMD_DELE: // unlink
			retval = dmlite_unlink(dmlite_context_obj, localpath);
			break;

		case GLOBUS_GFS_CMD_RNTO: // rename
			retval = dmlite_rename(dmlite_context_obj, dmlite_gfs_fixpath(cmd_info->from_pathname, GLOBUS_FALSE), localpath);
			break;

		case GLOBUS_GFS_CMD_CKSM: // compute chksum
			//TODO:
			break;

		case GLOBUS_GFS_CMD_SITE_CHMOD: // chmod
			retval = dmlite_chmod(dmlite_context_obj, localpath, cmd_info->chmod_mode);
			break;

		case GLOBUS_GFS_CMD_SITE_CHGRP: // chgrp
//			retval = dmlite_lchown(dmlite_context_obj, localpath,
//				-1, cmd_info->chgrp_group);  // TODO: proper group here
			break;

		default:
			result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, ENOTSUP, "unsupported command");
			goto errout;
	}

	if (retval) {
		result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
		goto errout;
	}

errout:
	if (dmlite_context_obj)
		dmlite_context_free(dmlite_context_obj);
	globus_gridftp_server_finished_command(op, result, GLOBUS_NULL); // TODO: give a proper error here
}

/*************************************************************************
 *  recv
 *  ----
 *  This interface function is called when the client requests that a
 *  file be transfered to the server.
 *
 *  To receive a file the following functions will be used in roughly
 *  the presented order.  They are doced in more detail with the
 *  gridftp server documentation.
 *
 *      globus_gridftp_server_begin_transfer();
 *      globus_gridftp_server_register_read();
 *      globus_gridftp_server_finished_transfer();
 *
 ************************************************************************/
static globus_bool_t globus_l_gfs_dmlite_recv_next_block(dmlite_handle_t * dmlite_handle_obj);

static void globus_l_gfs_dmlite_read_cb(
		globus_gfs_operation_t	op,
		globus_result_t		result,
		globus_byte_t *		buffer,
		globus_size_t		nbytes,
		globus_off_t		offset,
		globus_bool_t		eof,
		void *			user_arg)
{
	dmlite_handle_t *	dmlite_handle_obj = (dmlite_handle_t *)user_arg;
	struct dmlite_context *	dmlite_context_obj;
	globus_off_t		bytes_write;
	int			retval = 0;

	GlobusGFSName(globus_l_gfs_dmlite_read_cb);

	/* Convert the given handle */
	globus_mutex_lock(&dmlite_handle_obj->gfs_mutex);

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "read-cb: pending %d", dmlite_handle_obj->pending);
	/* We're done with one more write */
	dmlite_handle_obj->pending--;

	/* Are we done? If EOF is reached we are */
	if (eof == GLOBUS_TRUE)
		dmlite_handle_obj->done = GLOBUS_TRUE;

	if (result != GLOBUS_SUCCESS || dmlite_handle_obj->cur_length == 0 || nbytes == 0) { /* Nothing to write */
		dmlite_handle_obj->done = GLOBUS_TRUE;
		goto done;
	}

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "read-cb: ofs/len = %d/%d",
			dmlite_handle_obj->cur_offset + offset, dmlite_handle_obj->cur_length);

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "read-cb: got %d bytes at offset %d", nbytes, offset);
	/* Seek as we might be writing out-of-sequence */
	retval = dmlite_fseek(dmlite_handle_obj->fd, dmlite_handle_obj->cur_offset + offset, SEEK_SET);
	if (retval) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, retval, "failed to seek");
		dmlite_handle_obj->done = GLOBUS_TRUE;
		goto done;
	}

	/* Perform the actual write operation */
	bytes_write = dmlite_fwrite(dmlite_handle_obj->fd, buffer, nbytes);
	if (bytes_write < nbytes) { /* Check for errors */
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT, "failed write");
		dmlite_handle_obj->done = GLOBUS_TRUE;
		goto done;
	}
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "read-cb: wrote %d bytes", bytes_write);

	/* Update the remaining length to be written */
	if (dmlite_handle_obj->cur_length >= bytes_write) /* Never go below zero */
		dmlite_handle_obj->cur_length -= bytes_write;

done:
	globus_free(buffer);
	dmlite_handle_obj->cur_result = result;
	if (!dmlite_handle_obj->done)
		globus_l_gfs_dmlite_recv_next_block(dmlite_handle_obj);
	if (dmlite_handle_obj->pending != 0) {
		globus_mutex_unlock(&dmlite_handle_obj->gfs_mutex);
		return;
	}
	if ((dmlite_context_obj = dmlite_get_context(dmlite_handle_obj))) {
		if (dmlite_gfs_close(dmlite_context_obj, dmlite_handle_obj, result == GLOBUS_SUCCESS) &&
				result == GLOBUS_SUCCESS)
			dmlite_handle_obj->cur_result = dmlite_error2gfs_result(_gfs_name,
					dmlite_handle_obj, dmlite_context_obj);
		dmlite_context_free(dmlite_context_obj);
	} else {
		if (result == GLOBUS_SUCCESS) /* Do not report error here if we already handling one */
			dmlite_handle_obj->cur_result = posix_error2gfs_result(_gfs_name,
					dmlite_handle_obj, EFAULT, "failed to create context");
	}
	globus_gridftp_server_finished_transfer(op, dmlite_handle_obj->cur_result);
	globus_mutex_unlock(&dmlite_handle_obj->gfs_mutex);
}

static globus_bool_t globus_l_gfs_dmlite_recv_next_block(dmlite_handle_t * dmlite_handle_obj)
{
	globus_size_t		bytes_length;
	globus_byte_t *		buffer = NULL;
	struct dmlite_context *	dmlite_context_obj;

	GlobusGFSName(globus_l_gfs_dmlite_recv_next_block);

	/* Check we actually have something left to write */
	if (dmlite_handle_obj->cur_length == 0) {
		dmlite_handle_obj->cur_result = GLOBUS_SUCCESS;
		goto done;
	}

	/* How much should we actually write? block_size or less if bigger than remaining length */
	bytes_length = dmlite_handle_obj->block_size;
	if (dmlite_handle_obj->cur_length > 0 && dmlite_handle_obj->cur_length < dmlite_handle_obj->block_size)
		bytes_length = dmlite_handle_obj->cur_length;

	/* Allocate the write buffer */
	buffer = globus_malloc(bytes_length);
	if (!buffer) {
		dmlite_handle_obj->cur_result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj,
			EFAULT, "failed to allocate buffer of %d bytes", bytes_length);
		goto done;
	}
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "recv-next: buffer size is %d bytes", bytes_length);

	/* Register the read operation to push data to the client */
	dmlite_handle_obj->cur_result = globus_gridftp_server_register_read(dmlite_handle_obj->gfs_op, buffer,
			bytes_length, globus_l_gfs_dmlite_read_cb, dmlite_handle_obj);

	if (dmlite_handle_obj->cur_result == GLOBUS_SUCCESS) {
		dmlite_handle_obj->pending++;
		return GLOBUS_FALSE;
	}
	globus_free(buffer); /* Only in error case */

done:
	dmlite_handle_obj->done = GLOBUS_TRUE;
	if (dmlite_handle_obj->pending != 0)
		return GLOBUS_TRUE;
	if ((dmlite_context_obj = dmlite_get_context(dmlite_handle_obj))) {
		dmlite_gfs_close(dmlite_context_obj, dmlite_handle_obj,
				dmlite_handle_obj->cur_result == GLOBUS_SUCCESS);
		dmlite_context_free(dmlite_context_obj);
	}
	globus_gridftp_server_finished_transfer(dmlite_handle_obj->gfs_op, dmlite_handle_obj->cur_result);
	return GLOBUS_TRUE;
}

void globus_l_gfs_dmlite_recv(
		globus_gfs_operation_t		op,
		globus_gfs_transfer_info_t *	transfer_info,
		void *				user_arg)
{
	globus_bool_t		done = GLOBUS_FALSE;
	globus_result_t		result = GLOBUS_SUCCESS;
	dmlite_handle_t *	dmlite_handle_obj = (dmlite_handle_t *)user_arg;
	struct dmlite_context *	dmlite_context_obj;
	int			i;

	GlobusGFSName(globus_l_gfs_dmlite_recv);

	dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_DUMP, "recv: started");

	/* Get a dmlite_context object */
	dmlite_context_obj = dmlite_get_context(dmlite_handle_obj);
	if (!dmlite_context_obj) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT,
			"failed to create context");
		goto errout;
	}

	/* These two are values coming from globus */
	globus_gridftp_server_get_block_size(op, &dmlite_handle_obj->block_size);
	globus_gridftp_server_get_optimal_concurrency(op, &dmlite_handle_obj->optimal_count);

	/* Fill in the starting handle values for the write operations */
	dmlite_handle_obj->gfs_op = op;
	/* The rest are values using during the write sequence */
	dmlite_handle_obj->cur_result = GLOBUS_SUCCESS;
	dmlite_handle_obj->done = GLOBUS_FALSE;
	dmlite_handle_obj->pending = 0;
	/* These two in particular (offset/length) are set here when there are partial writes */
	globus_gridftp_server_get_write_range(dmlite_handle_obj->gfs_op,
		 &dmlite_handle_obj->cur_offset, &dmlite_handle_obj->cur_length);

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "recv: ofs/len = %d/%d",
			dmlite_handle_obj->cur_offset, dmlite_handle_obj->cur_length);

	/* Open the file and save the descriptor */
	dmlite_handle_obj->fd = dmlite_gfs_open(dmlite_context_obj, dmlite_handle_obj,
			transfer_info->pathname, O_WRONLY | O_CREAT);
	if (!dmlite_handle_obj->fd) {
		result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
		goto errout;
	}

	/* Initiate the write operations */
	globus_gridftp_server_begin_transfer(op, 0, dmlite_handle_obj);
	globus_mutex_lock(&dmlite_handle_obj->gfs_mutex);
	for(i = 0; i < dmlite_handle_obj->optimal_count && done != GLOBUS_TRUE; i++) {
		done = globus_l_gfs_dmlite_recv_next_block(dmlite_handle_obj);
	}
	globus_mutex_unlock(&dmlite_handle_obj->gfs_mutex);
	dmlite_context_free(dmlite_context_obj);
	return;

errout:
	if (dmlite_context_obj) {
		dmlite_gfs_close(dmlite_context_obj, dmlite_handle_obj, result == GLOBUS_SUCCESS);
		dmlite_context_free(dmlite_context_obj);
	}
	globus_gfs_log_result(GLOBUS_GFS_LOG_ERR,"recv: ", result);
	globus_gridftp_server_finished_transfer(op, result);
}

/*************************************************************************
 *  send
 *  ----
 *  This interface function is called when the client requests to receive
 *  a file from the server.
 *
 *  To send a file to the client the following functions will be used in roughly
 *  the presented order.  They are doced in more detail with the
 *  gridftp server documentation.
 *
 *      globus_gridftp_server_begin_transfer();
 *      globus_gridftp_server_register_write();
 *      globus_gridftp_server_finished_transfer();
 *
 ************************************************************************/
static globus_bool_t globus_l_gfs_dmlite_send_next_block(dmlite_handle_t * dmlite_handle_obj);

static void globus_l_gfs_dmlite_write_cb(
		globus_gfs_operation_t	op,
		globus_result_t		result,
		globus_byte_t *		buffer,
		globus_size_t		nbytes,
		void *			user_arg)
{
	dmlite_handle_t * dmlite_handle_obj = (dmlite_handle_t *)user_arg;

	GlobusGFSName(globus_l_gfs_dmlite_write_cb);

	globus_free(buffer);

	globus_mutex_lock(&dmlite_handle_obj->gfs_mutex);

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "write-cb: pending %d",
		dmlite_handle_obj->pending);

	/* We're done with one more write */
	dmlite_handle_obj->pending--;

	/* Are we done? If current result is not success we are */
	if (result != GLOBUS_SUCCESS) {
		dmlite_handle_obj->cur_result = result;
		dmlite_handle_obj->done = GLOBUS_TRUE;
	}
	if (!dmlite_handle_obj->done) {
		globus_l_gfs_dmlite_send_next_block(dmlite_handle_obj);
	}
	if (dmlite_handle_obj->pending == 0) {
		dmlite_gfs_close(NULL, dmlite_handle_obj, dmlite_handle_obj->cur_result == GLOBUS_SUCCESS);
		globus_gridftp_server_finished_transfer(op, dmlite_handle_obj->cur_result);
	}
	globus_mutex_unlock(&dmlite_handle_obj->gfs_mutex);
}

static globus_bool_t globus_l_gfs_dmlite_send_next_block(dmlite_handle_t * dmlite_handle_obj)
{
	globus_size_t		bytes_length;
	globus_off_t		bytes_read;
	globus_byte_t *		buffer = NULL;

	GlobusGFSName(globus_l_gfs_dmlite_send_next_block);

	/* Check we actually have something left to read */
	if (dmlite_handle_obj->cur_length == 0 || dmlite_feof(dmlite_handle_obj->fd))
		goto done;

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "send-next: ofs/len = %d/%d",
			dmlite_handle_obj->cur_offset, dmlite_handle_obj->cur_length);

	/* How much should we actually read? block_size or less if bigger than remaining length */
	bytes_length = dmlite_handle_obj->block_size;
	if (dmlite_handle_obj->cur_length > 0 && dmlite_handle_obj->cur_length < dmlite_handle_obj->block_size)
		bytes_length = dmlite_handle_obj->cur_length;

	/* Allocate the read buffer */
	buffer = globus_malloc(bytes_length);
	if (!buffer) {
		dmlite_handle_obj->cur_result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj,
				EFAULT, "failed to allocate buffer of %d bytes", bytes_length);
		goto done;
	}
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "send-next: buffer size is %d bytes", bytes_length);

	/* Perform the actual read operation */
	bytes_read = dmlite_fread(dmlite_handle_obj->fd, buffer, bytes_length);
	/* Check for end of file reached (0 bytes read) */

	if (bytes_read < 0) { /* Check for errors too */
		dmlite_handle_obj->cur_result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj,
				EFAULT, "failed read");
		goto done;
	}
	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "send-next: read %d bytes", bytes_read);

	/* Register the write operation to push data to the client */
	dmlite_handle_obj->cur_result = globus_gridftp_server_register_write(dmlite_handle_obj->gfs_op,
			buffer, bytes_read, dmlite_handle_obj->cur_offset, -1,
			globus_l_gfs_dmlite_write_cb, dmlite_handle_obj);

	if (dmlite_handle_obj->cur_result != GLOBUS_SUCCESS)
		goto done;

	/* Update the current read offset and remaining length to be read */
	dmlite_handle_obj->pending++;
	dmlite_handle_obj->cur_offset += bytes_read;
	if (dmlite_handle_obj->cur_length >= bytes_read) /* Never go below zero */
		dmlite_handle_obj->cur_length -= bytes_read;

	return GLOBUS_FALSE;

done:
	if (buffer)
		globus_free(buffer);
	dmlite_handle_obj->done = GLOBUS_TRUE;
	if (dmlite_handle_obj->pending == 0) {
		dmlite_gfs_close(NULL, dmlite_handle_obj, dmlite_handle_obj->cur_result == GLOBUS_SUCCESS);
		globus_gridftp_server_finished_transfer(dmlite_handle_obj->gfs_op, dmlite_handle_obj->cur_result);
	}
	return GLOBUS_TRUE;
}

void globus_l_gfs_dmlite_send(
		globus_gfs_operation_t		op,
		globus_gfs_transfer_info_t *	transfer_info,
		void *				user_arg)
{
	globus_bool_t		done = GLOBUS_FALSE;
	globus_result_t		result = GLOBUS_SUCCESS;
	dmlite_handle_t *	dmlite_handle_obj = (dmlite_handle_t *)user_arg;
	struct dmlite_context *	dmlite_context_obj;
	int			i;
	int			retval = 0;

	GlobusGFSName(globus_l_gfs_dmlite_send);

	dmlite_gfs_log(dmlite_handle_obj, GLOBUS_GFS_LOG_DUMP, "send: started");

	/* Get a dmlite_context object */
	dmlite_context_obj = dmlite_get_context(dmlite_handle_obj);
	if (!dmlite_context_obj) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, EFAULT,
			"failed to create context");
		goto errout;
	}

	/* These two are values coming from globus */
	globus_gridftp_server_get_block_size(op, &dmlite_handle_obj->block_size);
	globus_gridftp_server_get_optimal_concurrency(op, &dmlite_handle_obj->optimal_count);

	/* Fill in the starting handle values for the read operations */
	dmlite_handle_obj->gfs_op = op;
	/* The rest are values using during the read sequence */
	dmlite_handle_obj->cur_result = GLOBUS_SUCCESS;
	dmlite_handle_obj->done = GLOBUS_FALSE;
	dmlite_handle_obj->pending = 0;
	/* These two in particular (offset/length) are set here when there are partial reads */
	globus_gridftp_server_get_read_range(dmlite_handle_obj->gfs_op,
		 &dmlite_handle_obj->cur_offset, &dmlite_handle_obj->cur_length);

	dmlite_gfs_log(NULL, GLOBUS_GFS_LOG_DUMP, "send: ofs/len = %d/%d",
			dmlite_handle_obj->cur_offset, dmlite_handle_obj->cur_length);

	/* Open the file and save the descriptor */
	dmlite_handle_obj->fd = dmlite_gfs_open(dmlite_context_obj, dmlite_handle_obj,
			transfer_info->pathname, O_RDONLY);
	if (!dmlite_handle_obj->fd) {
		result = dmlite_error2gfs_result(_gfs_name, dmlite_handle_obj, dmlite_context_obj);
		goto errout;
	}

	/* Seek as we might be reading partially */
	retval = dmlite_fseek(dmlite_handle_obj->fd, dmlite_handle_obj->cur_offset, SEEK_SET);
	if (retval) {
		result = posix_error2gfs_result(_gfs_name, dmlite_handle_obj, retval, "failed to seek");
		goto errout;
	}

	/* Initiate the read operations */
	globus_gridftp_server_begin_transfer(op, 0, dmlite_handle_obj);
	globus_mutex_lock(&dmlite_handle_obj->gfs_mutex);
	for(i = 0; i < dmlite_handle_obj->optimal_count && done != GLOBUS_TRUE; i++) {
		done = globus_l_gfs_dmlite_send_next_block(dmlite_handle_obj);
	}
	globus_mutex_unlock(&dmlite_handle_obj->gfs_mutex);
	dmlite_context_free(dmlite_context_obj);
	return;

errout:
	if (dmlite_handle_obj->fd)
		dmlite_gfs_close(NULL, dmlite_handle_obj, result == GLOBUS_SUCCESS);
	if (dmlite_context_obj)
		dmlite_context_free(dmlite_context_obj);
	globus_gfs_log_result(GLOBUS_GFS_LOG_ERR, "send: ", result);
	globus_gridftp_server_finished_transfer(op, result);
}
